from __future__ import division

import os
import numpy as np
from scipy.spatial.distance import cdist
from scipy import stats
from scipy.special import gamma
from .mathUtils import integrationUtils
from . import Sed, Bandpass, BandpassDict, spectra
from .lsst_utils import calcMagError_sed, calcSNR_sed
from .lsst_utils import PhotometricParameters

__all__ = ["calcIG"]


class calcIG(integrationUtils):

    """
    This class will take a set of SEDs and a set of filters
    and calculate the information gain.
    """

    def __init__(self, filter_dict, sed_list, sed_probs, sed_mags=22.0,
                 sky_mag=19.0, ref_filter=None, phot_params=None,
                 fwhm_eff=1.0):

        self._sed_list = []

        if ref_filter is None:
            ref_filter = Bandpass()
            ref_filter.imsimBandpass()

        for sed_obj in sed_list:

            sed_copy = Sed()
            sed_copy.setSED(wavelen=sed_obj.wavelen,
                            flambda=sed_obj.flambda)
            sed_copy.resampleSED(wavelen_match=filter_dict.values()[0].wavelen)
            sed_copy.flambda[np.where(np.isnan(sed_copy.flambda))] = 0.

            f_norm = sed_copy.calcFluxNorm(sed_mags, ref_filter)
            sed_copy.multiplyFluxNorm(f_norm)
            self._sed_list.append(sed_copy)

        self._hardware_filt_dict, self._total_filt_dict = \
            BandpassDict.addSystemBandpass(filter_dict)

        self.sky_spec = spectra().get_dark_sky_spectrum()
        sky_fn = self.sky_spec.calcFluxNorm(sky_mag, ref_filter)
        self.sky_spec.multiplyFluxNorm(sky_fn)

        if phot_params is None:
            # Use Default LSST parameters
            self.phot_params = PhotometricParameters()
        else:
            self.phot_params = phot_params

        self.sed_probs = np.array(sed_probs)/np.sum(sed_probs)
        self.fwhm_eff = fwhm_eff

        return

    def calc_colors(self, return_all=False):

        """
        Calculate the colors and errors in color measurement
        of the SEDs in the given filters.
        """

        sed_colors = []
        color_errors = []
        snr_values = []
        sed_mag_list = []

        sky_mags = self._total_filt_dict.magListForSed(self.sky_spec)

        for sed_obj in self._sed_list:

            sed_mags = self._total_filt_dict.magListForSed(sed_obj)

            mag_errors = [calcMagError_sed(sed_obj, filt_tot,
                                           self.sky_spec, filt_hw,
                                           self.phot_params, self.fwhm_eff) for
                          filt_tot, filt_hw in zip(
                                self._total_filt_dict.values(),
                                self._hardware_filt_dict.values())]

            if np.isnan(sed_mags[0]):
                print(sed_mags)
            sed_colors.append([sed_mags[i] - sed_mags[i+1]
                               for i in range(len(sed_mags) - 1)])
            color_errors.append([np.sqrt(mag_errors[i]**2. +
                                         mag_errors[i+1]**2.) for i
                                 in range(len(mag_errors) - 1)])

            if return_all is True:
                snr_value = [calcSNR_sed(sed_obj, filt_tot,
                                         self.sky_spec, filt_hw,
                                         self.phot_params, self.fwhm_eff) for
                             filt_tot, filt_hw in zip(
                                self._total_filt_dict.values(),
                                self._hardware_filt_dict.values())]
                snr_values.append(snr_value)
                sed_mag_list.append(sed_mags)

        if return_all is True:
            return np.array(sed_colors), np.array(color_errors), snr_values,\
                   sed_mag_list, sky_mags

        return np.array(sed_colors), np.array(color_errors)

    def calc_h(self):

        """
        Calculates the total entropy of the set of SEDs and redshifts.
        """

        sed_probs = self.sed_probs

        h_sum = 0
        total_y = len(sed_probs)

        for py_i in sed_probs:
            h_sum += -(py_i)*np.log2(py_i)

        return h_sum

    def calc_hyx(self, colors, errors, rand_state=None):

        """
        Calculate the conditional entropy.
        """

        if rand_state is None:
            rand_state = np.random.RandomState()

        sed_probs = self.sed_probs

        hyx_sum = 0

        num_seds, num_colors = np.shape(colors)

        rv = rand_state.multivariate_normal

        y_vals = []
        y_distances = []
        num_points = 10000
        x_total = np.zeros((num_seds*num_points, num_colors))

        for idx in range(num_seds):

            y_samples = rv(mean=colors[idx],
                           cov=np.diagflat(errors[idx]**2.),
                           size=num_points)

            y_samples = y_samples.reshape(num_points, num_colors)

            inv_cov = np.linalg.inv(np.diagflat(errors[idx]**2.))

            y_dist = cdist(y_samples, [colors[idx]], metric='mahalanobis',
                           VI=inv_cov).flatten()
            y_sort = np.argsort(y_dist)
            y_dist = y_dist[y_sort]
            y_samples = y_samples[y_sort]

            x_total[idx*num_points:(idx+1)*num_points] = \
                y_samples

            y_vals.append(y_samples)
            y_distances.append(y_dist)

        y_vals = np.array(y_vals)

        x_dens = np.zeros(len(x_total))

        for idx in range(num_seds):
            pdf_dist = stats.multivariate_normal
            y_dens = sed_probs[idx] * \
                pdf_dist.pdf(x_total, mean=colors[idx],
                             cov=np.diagflat(errors[idx])**2.)
            x_dens += y_dens

        for idx in range(num_seds):

            y_samp = x_total[idx*num_points:(idx+1)*num_points]
            y_dens = sed_probs[idx] * \
                pdf_dist.pdf(y_samp, mean=colors[idx],
                             cov=np.diagflat(errors[idx])**2.)

            norm_factor = self.calc_integral_scaling(y_distances[idx],
                                                     num_colors, errors[idx])

            hyx_i = np.nansum(norm_factor * (y_dens * np.log2(y_dens /
                                             x_dens[idx*num_points:(idx+1) *
                                                    num_points])))
            hyx_sum += hyx_i

        return -1.*hyx_sum

    def calc_IG(self, rand_state=None):

        """
        Use the entropy and conditional entropy methods and then subtract
        to get information gain.
        """

        colors, errors = self.calc_colors()
        hy_sum = self.calc_h()
        hyx_sum = self.calc_hyx(colors, errors, rand_state)

        info_gain = hy_sum - hyx_sum

        return info_gain



import numpy as np
from scipy.special import gamma

__all__ = ["integrationUtils"]


class integrationUtils(object):

    def __init__(self):

        return

    def calc_integral_scaling(self, distances, n_dim, dim_scale):

        """
        The integration works by calculating the volume of
        slices of a hyperellipsoid under the probability density
        curve.

        This part calculates the area of each slice and when
        it is returned it is multiplied by the pdf value
        to get a volume.

        Since the width of the probability distribution
        in each dimension will be different this scales
        each dimension in the integral appropriately.
        """

        # (r_outer ^ d) - (r_inner ^ d)
        norm_factor = (distances[1:]**n_dim -
                       distances[:-1]**n_dim)
        norm_factor = np.append((distances[0]**n_dim),
                                norm_factor)

        # Volume of hyperellipsoid: https://math.stackexchange.com/
        # questions/332391/volume-of-hyperellipsoid with V_d term
        # from wikipedia: https://en.wikipedia.org/wiki/N-sphere

        norm_factor *= ((np.pi**(n_dim/2.))/gamma((n_dim/2.)+1))
        norm_factor *= np.linalg.det(np.diag(dim_scale))

        return norm_factor



# pylint: skip-file
#
# LSST Data Management System
# Copyright 2008, 2009, 2010, 2011, 2012 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

"""
bandpass -
Class data:
 wavelen (nm)
 sb  (Transmission, 0-1)
 phi (Normalized system response)
  wavelen/sb are guaranteed gridded.
  phi will be None until specifically needed;
     any updates to wavelen/sb within class will reset phi to None.
 the name of the bandpass file
Note that Bandpass objects are required to maintain a uniform grid in wavelength, rather than
being allowed to have variable wavelength bins. This is because of the method used in 'Sed' to
calculate magnitudes, but is simpler to enforce here.
Methods:
 __init__ : pass wavelen/sb arrays and set values (on grid) OR set data to None's
 setWavelenLimits / getWavelenLimits: set or get the wavelength limits of bandpass
 setBandpass: set bandpass using wavelen/sb input values
 getBandpass: return copies of wavelen/sb values
 imsimBandpass : set up a bandpass which is 0 everywhere but one wavelength
                 (this can be useful for imsim magnitudes)
 readThroughput : set up a bandpass by reading data from a single file
 readThroughtputList : set up a bandpass by reading data from many files and multiplying
                       the individual throughputs
 resampleBandpass : use linear interpolation to resample wavelen/sb arrays onto a regular grid
                    (grid is specified by min/max/step size)
 sbTophi : calculate phi from sb - needed for calculating magnitudes
 multiplyThroughputs : multiply self.wavelen/sb by given wavelen/sb and return
                       new wavelen/sb arrays (gridded like self)
 calcZP_t : calculate instrumental zeropoint for this bandpass
 calcEffWavelen: calculate the effective wavelength (using both Sb and Phi) for this bandpass
 writeThroughput : utility to write bandpass information to file
"""
from __future__ import print_function
from builtins import range
from builtins import object
import os
import warnings
import numpy
import scipy.interpolate as interpolate
import gzip
from .PhysicalParameters import PhysicalParameters
from .Sed import Sed  # For ZP_t and M5 calculations. And for 'fast mags' calculation.

__all__ = ["Bandpass"]


class Bandpass(object):
    """
    Class for holding and utilizing telescope bandpasses.
    """
    def __init__(self, wavelen=None, sb=None,
                 wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Initialize bandpass object, with option to pass wavelen/sb arrays in directly.
        Also can specify wavelength grid min/max/step or use default - sb and wavelen will
        be resampled to this grid. If wavelen/sb are given, these will be set, but phi
        will be set to None.
        Otherwise all set to None and user should call readThroughput, readThroughputList,
        or imsimBandpass to populate bandpass data.
        """

        self._physParams = PhysicalParameters()

        if wavelen_min is None:
            if wavelen is None:
                wavelen_min = self._physParams.minwavelen
            else:
                wavelen_min = wavelen.min()
        if wavelen_max is None:
            if wavelen is None:
                wavelen_max = self._physParams.maxwavelen
            else:
                wavelen_max = wavelen.max()
        if wavelen_step is None:
            if wavelen is None:
                wavelen_step = self._physParams.wavelenstep
            else:
                wavelen_step = numpy.diff(wavelen).min()
        self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        self.wavelen=None
        self.sb=None
        self.phi=None
        self.bandpassname = None
        if (wavelen is not None) and (sb is not None):
            self.setBandpass(wavelen, sb, wavelen_min, wavelen_max, wavelen_step)

        return

    ## getters and setters
    def setWavelenLimits(self, wavelen_min, wavelen_max, wavelen_step):
        """
        Set internal records of wavelen limits, _min, _max, _step.
        """
        # If we've been given values for wavelen_min, _max, _step, set them here.
        if wavelen_min is not None:
            self.wavelen_min = wavelen_min
        if wavelen_max is not None:
            self.wavelen_max = wavelen_max
        if wavelen_step is not None:
            self.wavelen_step = wavelen_step
        return

    def getWavelenLimits(self, wavelen_min, wavelen_max, wavelen_step):
        """
        Return appropriate wavelen limits (_min, _max, _step) if passed None values.
        """
        if wavelen_min is None:
            wavelen_min = self.wavelen_min
        if wavelen_max is None:
            wavelen_max = self.wavelen_max
        if wavelen_step is None:
            wavelen_step = self.wavelen_step
        return wavelen_min, wavelen_max, wavelen_step

    def setBandpass(self, wavelen, sb,
                    wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Populate bandpass data with wavelen/sb arrays.
        Sets self.wavelen/sb on a grid of wavelen_min/max/step. Phi set to None.
        """
        self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Check data type.
        if (isinstance(wavelen, numpy.ndarray)==False) or (isinstance(sb, numpy.ndarray)==False):
            raise ValueError("Wavelen and sb arrays must be numpy arrays.")
        # Check data matches in length.
        if (len(wavelen)!=len(sb)):
                raise ValueError("Wavelen and sb arrays must have the same length.")
        # Data seems ok then, make a new copy of this data for self.
        self.wavelen = numpy.copy(wavelen)
        self.phi = None
        self.sb = numpy.copy(sb)
        # Resample wavelen/sb onto grid.
        self.resampleBandpass(wavelen_min=wavelen_min, wavelen_max=wavelen_max, wavelen_step=wavelen_step)
        self.bandpassname = 'FromArrays'
        return

    def imsimBandpass(self, imsimwavelen=500.0,
                      wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Populate bandpass data with sb=0 everywhere except sb=1 at imsimwavelen.
        Sets wavelen/sb, with grid min/max/step as optional parameters. Does NOT set phi.
        """
        self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Set up arrays.
        self.wavelen = numpy.arange(self.wavelen_min, self.wavelen_max+self.wavelen_step,
                                    self.wavelen_step, dtype='float')
        self.phi = None
        # Set sb.
        self.sb = numpy.zeros(len(self.wavelen), dtype='float')
        self.sb[abs(self.wavelen-imsimwavelen)<self.wavelen_step/2.0] = 1.0
        self.bandpassname = 'IMSIM'
        return

    def readThroughput(self, filename, wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Populate bandpass data with data (wavelen/sb) read from file, resample onto grid.
        Sets wavelen/sb, with grid min/max/step as optional parameters. Does NOT set phi.
        """
        self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Set self values to None in case of file read error.
        self.wavelen = None
        self.phi = None
        self.sb = None
        # Check for filename error.
        # If given list of filenames, pass to (and return from) readThroughputList.
        if isinstance(filename, list):
            warnings.warn("Was given list of files, instead of a single file. Using readThroughputList instead")
            self.readThroughputList(componentList=filename,
                                    wavelen_min=self.wavelen_min, wavelen_max=self.wavelen_max,
                                    wavelen_step=self.wavelen_step)
        # Filename is single file, now try to open file and read data.
        try:
            if filename.endswith('.gz'):
                f = gzip.open(filename, 'r')
            else:
                print(filename)
                f = open(filename, 'r')
        except IOError:
            try:
                if filename.endswith('.gz'):
                    f = open(filename[:-3], 'r')
                else:
                    f = gzip.open(filename+'.gz', 'r')
            except IOError:
                raise IOError('The throughput file %s does not exist' %(filename))
        # The throughput file should have wavelength(A), throughput(Sb) as first two columns.
        wavelen = []
        sb = []
        for line in f:
            print(line)
            if line.startswith("#") or line.startswith('$') or line.startswith('!'):
                continue
            values = line.split()
            if len(values)<2:
                continue
            if (values[0] == '$') or (values[0] =='#') or (values[0] =='!'):
                continue
            wavelen.append(float(values[0]))
            sb.append(float(values[1]))
        f.close()
        self.bandpassname = filename
        # Set up wavelen/sb.
        self.wavelen = numpy.array(wavelen, dtype='float')
        self.sb = numpy.array(sb, dtype='float')
        # Check that wavelength is monotonic increasing and non-repeating in wavelength. (Sort on wavelength).
        if len(self.wavelen) != len(numpy.unique(self.wavelen)):
            raise ValueError('The wavelength values in file %s are non-unique.' %(filename))
        # Sort values.
        p = self.wavelen.argsort()
        self.wavelen = self.wavelen[p]
        self.sb = self.sb[p]
        # Resample throughput onto grid.
        if self.needResample():
            self.resampleBandpass()
        if self.sb.sum() < 1e-300:
            raise Exception("Bandpass data from %s has no throughput in desired grid range %f, %f" %(filename, wavelen_min, wavelen_max))
        return

    def readThroughputList(self, componentList=['detector.dat', 'lens1.dat',
                                                'lens2.dat', 'lens3.dat',
                                                'm1.dat', 'm2.dat', 'm3.dat',
                                                'atmos_std.dat'],
                           rootDir = '.',
                           wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Populate bandpass data by reading from a series of files with wavelen/Sb data.
        Multiplies throughputs (sb) from each file to give a final bandpass throughput.
        Sets wavelen/sb, with grid min/max/step as optional parameters.  Does NOT set phi.
        """
        # ComponentList = names of files in that directory.
        # A typical component list of all files to build final component list, including filter, might be:
        #   componentList=['detector.dat', 'lens1.dat', 'lens2.dat', 'lens3.dat',
        #                 'm1.dat', 'm2.dat', 'm3.dat', 'atmos_std.dat', 'ideal_g.dat']
        #
        # Set wavelen limits for this object, if any updates have been given.
        self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Set up wavelen/sb on grid.
        self.wavelen = numpy.arange(self.wavelen_min, self.wavelen_max+self.wavelen_step/2., self.wavelen_step,
                                    dtype='float')
        self.phi = None
        self.sb = numpy.ones(len(self.wavelen), dtype='float')
        # Set up a temporary bandpass object to hold data from each file.
        tempbandpass = Bandpass(wavelen_min=self.wavelen_min, wavelen_max=self.wavelen_max,
                                wavelen_step=self.wavelen_step)
        for component in componentList:
            # Read data from file.
            tempbandpass.readThroughput(os.path.join(rootDir, component))
            # Multiply self by new sb values.
            self.sb = self.sb * tempbandpass.sb
        self.bandpassname = ''.join(componentList)
        return

    def getBandpass(self):
        wavelen = numpy.copy(self.wavelen)
        sb = numpy.copy(self.sb)
        return wavelen, sb

    ## utilities

    def checkUseSelf(self, wavelen, sb):
        """
        Simple utility to check if should be using self.wavelen/sb or passed arrays.
        Useful for other methods in this class.
        Also does data integrity check on wavelen/sb if not self.
        """
        update_self = False
        if (wavelen is None) or (sb is None):
            # Then one of the arrays was not passed - check if true for both.
            if (wavelen is not None) or (sb is not None):
                # Then only one of the arrays was passed - raise exception.
                raise ValueError("Must either pass *both* wavelen/sb pair, or use self defaults")
            # Okay, neither wavelen or sb was passed in - using self only.
            update_self = True
        else:
            # Both of the arrays were passed in - check their validity.
            if (isinstance(wavelen, numpy.ndarray)==False) or (isinstance(sb, numpy.ndarray)==False):
                raise ValueError("Must pass wavelen/sb as numpy arrays")
            if len(wavelen)!=len(sb):
                raise ValueError("Must pass equal length wavelen/sb arrays")
        return update_self

    def needResample(self, wavelen=None,
                     wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Return true/false of whether wavelen need to be resampled onto a grid.
        Given wavelen OR defaults to self.wavelen/sb - return True/False check on whether
        the arrays need to be resampled to match wavelen_min/max/step grid.
        """
        # Thought about adding wavelen_match option here (to give this an array to match to, rather than
        # the grid parameters .. but then thought bandpass always needs to be on a regular grid (because
        # of magnitude calculations). So, this will stay match to the grid parameters only.
        # Check wavelength limits.
        wavelen_min, wavelen_max, wavelen_step = self.getWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Check if method acting on self or other data (here, using data type checks primarily).
        update_self = self.checkUseSelf(wavelen, wavelen)
        if update_self:
            wavelen = self.wavelen
        wavelen_max_in = wavelen[len(wavelen)-1]
        wavelen_min_in = wavelen[0]
        wavelen_step_in = wavelen[1]-wavelen[0]
        # Start check if data is already gridded.
        need_regrid=True
        # First check minimum/maximum and first step in array.
        if ((wavelen_min_in == wavelen_min) and (wavelen_max_in == wavelen_max)):
            # Then check on step size.
            stepsize = numpy.unique(numpy.diff(wavelen))
            if (len(stepsize) == 1) and (stepsize[0] == wavelen_step):
                need_regrid = False
        # At this point, need_grid=True unless it's proven to be False, so return value.
        return need_regrid

    def resampleBandpass(self, wavelen=None, sb=None,
                         wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Resamples wavelen/sb (or self.wavelen/sb) onto grid defined by min/max/step.
        Either returns wavelen/sb (if given those arrays) or updates wavelen / Sb in self.
        If updating self, resets phi to None.
        """
        # Check wavelength limits.
        wavelen_min, wavelen_max, wavelen_step = self.getWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
        # Is method acting on self.wavelen/sb or passed in wavelen/sb? Sort it out.
        update_self = self.checkUseSelf(wavelen, sb)
        if update_self:
            wavelen = self.wavelen
            sb = self.sb
        # Now, on with the resampling.
        if (wavelen.min() > wavelen_max) or (wavelen.max() < wavelen_min):
            raise Exception("No overlap between known wavelength range and desired wavelength range.")
        # Set up gridded wavelength.
        wavelen_grid = numpy.arange(wavelen_min, wavelen_max+wavelen_step/2.0, wavelen_step, dtype='float')
        # Do the interpolation of wavelen/sb onto the grid. (note wavelen/sb type failures will die here).
        f = interpolate.interp1d(wavelen, sb, fill_value=0, bounds_error=False)
        sb_grid = f(wavelen_grid)
        # Update self values if necessary.
        if update_self:
            self.phi = None
            self.wavelen = wavelen_grid
            self.sb = sb_grid
            self.setWavelenLimits(wavelen_min, wavelen_max, wavelen_step)
            return
        return wavelen_grid, sb_grid

    ## more complicated bandpass functions

    def sbTophi(self):
        """
        Calculate and set phi - the normalized system response.
        This function only pdates self.phi.
        """
        # The definition of phi = (Sb/wavelength)/\int(Sb/wavelength)dlambda.
        # Due to definition of class, self.sb and self.wavelen are guaranteed equal-gridded.
        dlambda = self.wavelen[1]-self.wavelen[0]
        self.phi = self.sb/self.wavelen
        # Normalize phi so that the integral of phi is 1.
        phisum = self.phi.sum()
        if phisum < 1e-300:
            raise Exception("Phi is poorly defined (nearly 0) over bandpass range.")
        norm = phisum * dlambda
        self.phi = self.phi / norm
        return

    def multiplyThroughputs(self, wavelen_other, sb_other):
        """
        Multiply self.sb by another wavelen/sb pair, return wavelen/sb arrays.
        The returned arrays will be gridded like this bandpass.
        This method does not affect self.
        """
        # Resample wavelen_other/sb_other to match this bandpass.
        if self.needResample(wavelen=wavelen_other):
            wavelen_other, sb_other = self.resampleBandpass(wavelen=wavelen_other, sb=sb_other)
        # Make new memory copy of wavelen.
        wavelen_new = numpy.copy(self.wavelen)
        # Calculate new transmission - this is also new memory.
        sb_new = self.sb * sb_other
        return wavelen_new, sb_new

    def calcZP_t(self, photometricParameters):
        """
        Calculate the instrumental zeropoint for a bandpass.
        @param [in] photometricParameters is an instantiation of the
        PhotometricParameters class that carries details about the
        photometric response of the telescope.  Defaults to LSST values.
        """
        # ZP_t is the magnitude of a (F_nu flat) source which produced 1 count per second.
        # This is often also known as the 'instrumental zeropoint'.
        # Set gain to 1 if want to explore photo-electrons rather than adu.
        # The typical LSST exposure time is 15s and this is default here, but typical zp_t definition is for 1s.
        # SED class uses flambda in ergs/cm^2/s/nm, so need effarea in cm^2.
        #
        # Check dlambda value for integral.
        dlambda = self.wavelen[1] - self.wavelen[0]
        # Set up flat source of arbitrary brightness,
        #   but where the units of fnu are Jansky (for AB mag zeropoint = -8.9).
        flatsource = Sed()
        flatsource.setFlatSED(wavelen_min=self.wavelen_min, wavelen_max=self.wavelen_max,
                              wavelen_step=self.wavelen_step)
        adu = flatsource.calcADU(self, photParams=photometricParameters)
        # Scale fnu so that adu is 1 count/expTime.
        flatsource.fnu = flatsource.fnu * (1/adu)
        # Now need to calculate AB magnitude of the source with this fnu.
        if self.phi is None:
            self.sbTophi()
        zp_t = flatsource.calcMag(self)
        return zp_t


    def calcEffWavelen(self):
        """
        Calculate effective wavelengths for filters.
        """
        # This is useful for summary numbers for filters.
        # Calculate effective wavelength of filters.
        if self.phi is None:
            self.sbTophi()
        effwavelenphi = (self.wavelen*self.phi).sum()/self.phi.sum()
        effwavelensb = (self.wavelen*self.sb).sum()/self.sb.sum()
        return effwavelenphi, effwavelensb


    def writeThroughput(self, filename, print_header=None, write_phi=False):
        """
        Write throughput to a file.
        """
        # Useful if you build a throughput up from components and need to record the combined value.
        f = open(filename, 'w')
        # Print header.
        if print_header is not None:
            if not print_header.startswith('#'):
                print_header = '#' + print_header
            f.write(print_header)
        if write_phi:
            if self.phi is None:
                self.sbTophi()
            print("# Wavelength(nm)  Throughput(0-1)   Phi", file=f)
        else:
            print("# Wavelength(nm)  Throughput(0-1)", file=f)
        # Loop through data, printing out to file.
        for i in range(0, len(self.wavelen), 1):
            if write_phi:
                print(self.wavelen[i], self.sb[i], self.phi[i], file=f)
            else:
                print(self.wavelen[i], self.sb[i], file=f)
        f.close()
        return



import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from scipy import stats
from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib import tri
from . import filters, calcIG, _siggiBase
from .lsst_utils import BandpassDict

__all__ = ["plotting"]


class plotting(_siggiBase):

    def __init__(self, sed_list, best_point, set_ratio=None,
                 frozen_filt_dict=None, frozen_filt_eff_wavelen=None,
                 sky_mag=19.0, sed_mags=22.0):

        f = filters()

        if set_ratio is not None:

            filter_info = []

            for i in range(int(len(best_point)/2)):
                edges = np.array(best_point[2*i:2*(i+1)])
                bottom_len = edges[1] - edges[0]
                top_len = set_ratio*bottom_len
                center = edges[0] + bottom_len/2.
                top_left = center - top_len/2.
                top_right = center + top_len/2.
                filter_info.append([edges[0], top_left, top_right, edges[1]])
        else:
            filter_info = [best_point[4*i:4*(i+1)]
                           for i in range(int(len(best_point)/4))]

        trap_dict = f.trap_filters(filter_info)

        hardware_filt_dict, total_filt_dict = \
            BandpassDict.addSystemBandpass(trap_dict)

        if frozen_filt_dict is None:
            self.filter_dict = total_filt_dict
        else:
            if (type(frozen_filt_eff_wavelen) != list):
                raise ValueError("If including frozen filters, " +
                                 "need list of eff. wavelengths.")
            filter_wavelengths = frozen_filt_eff_wavelen +\
                self.find_filt_centers(filter_info)
            filter_names_unsort = frozen_filt_dict.keys() +\
                total_filt_dict.keys()
            filter_list_unsort = frozen_filt_dict.values() +\
                total_filt_dict.values()
            sort_idx = np.argsort(filter_wavelengths)
            filter_names = [filter_names_unsort[idx] for idx in sort_idx]
            filter_list = [filter_list_unsort[idx] for idx in sort_idx]

            self.filter_dict = BandpassDict(filter_list, filter_names)

        self.sed_list = sed_list

        self.sky_mag = sky_mag
        self.sed_mags = sed_mags
        self.set_ratio = set_ratio

    def plot_filters(self, fig=None):

        """
        Plot the filters over the top of the SEDs used to optimize.
        """

        if fig is None:
            fig = plt.figure(figsize=(12, 6))

        for sed_obj in self.sed_list:
            plt.plot(sed_obj.wavelen, sed_obj.flambda/np.max(sed_obj.flambda),
                     c='k', alpha=0.5)

        c_list = np.linspace(0, 1, len(self.filter_dict.values()))

        cmap = plt.get_cmap('rainbow')

        for name, filt, color in zip(self.filter_dict.keys(),
                                     self.filter_dict.values(), c_list):
            plt.fill(filt.wavelen, filt.sb,
                     c=cmap(color),
                     zorder=10, alpha=0.6, label=name)
        plt.xlim(filt.wavelen[0] - 50., filt.wavelen[-1] + 50)
        plt.xlabel('Wavelength (nm)')
        plt.ylabel('Scaled Flux')

        return fig

    # The following come from the ipython notebook here:
    # http://nbviewer.jupyter.org/github/dpsanders/matplotlib-examples/blob/master/colorline.ipynb

    # Data manipulation:

    def make_segments(self, x, y):
        '''
        Create list of line segments from x and y coordinates,
        in the correct format for LineCollection:
        an array of the form   numlines x (points per line) x 2 (x and y) array
        '''

        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        return segments

    # Interface to LineCollection:

    def colorline(self, x, y, z=None, cmap=plt.get_cmap('copper'),
                  norm=plt.Normalize(0.0, 1.0), linewidth=3, alpha=1.0):
        '''
        Plot a colored line with coordinates x and y
        Optionally specify colors in the array z
        Optionally specify a colormap, a norm function and a line width
        '''

        # Default colors equally spaced on [0,1]:
        if z is None:
            z = np.linspace(0.0, 1.0, len(x))

        # Special case if a single number:
        # to check for numerical input -- this is a hack
        if not hasattr(z, "__iter__"):
            z = np.array([z])

        z = np.asarray(z)

        segments = self.make_segments(x, y)
        lc = LineCollection(segments, array=z, cmap=cmap, norm=norm,
                            linewidth=linewidth, alpha=alpha)

        ax = plt.gca()
        ax.add_collection(lc)

        return lc

    def plot_color_color(self, filter_names, redshift_list,
                         cmap=plt.get_cmap('plasma'), fig=None,
                         include_err=True):

        """
        Plot the color-color tracks for each SED template as a function
        of redshift using pairs of filters.
        """

        if fig is None:
            fig = plt.figure(figsize=(12, 6))

        shift_seds = []

        for sed_obj in self.sed_list:
            for z_val in redshift_list:
                sed_copy = deepcopy(sed_obj)
                sed_copy.redshiftSED(z_val)
                shift_seds.append(sed_copy)

        color_x_dict = BandpassDict([self.filter_dict[filt] for filt
                                     in filter_names[:2]], filter_names[:2])
        color_y_dict = BandpassDict([self.filter_dict[filt] for filt
                                     in filter_names[2:]], filter_names[2:])

        calc_ig = calcIG(color_x_dict, shift_seds,
                         np.ones(len(shift_seds)),
                         sky_mag=self.sky_mag, sed_mags=self.sed_mags)
        col_x, err_x = calc_ig.calc_colors()

        calc_ig = calcIG(color_y_dict, shift_seds,
                         np.ones(len(shift_seds)),
                         sky_mag=self.sky_mag, sed_mags=self.sed_mags)
        col_y, err_y = calc_ig.calc_colors()

        num_z = len(redshift_list)

        for sed_num in range(len(self.sed_list)):

            start_idx = sed_num*num_z
            end_idx = start_idx + num_z

            self.colorline(col_x[start_idx:end_idx],
                           col_y[start_idx:end_idx],
                           cmap=cmap)

        plt.xlabel('%s - %s' % (filter_names[0], filter_names[1]))
        plt.ylabel('%s - %s' % (filter_names[2], filter_names[3]))

        plt.xlim(np.min(col_x) - 0.5, np.max(col_x) + 0.5)
        plt.ylim(np.min(col_y) - 0.5, np.max(col_y) + 0.5)

        sm = plt.cm.ScalarMappable(cmap=cmap,
                                   norm=plt.Normalize(
                                       vmin=np.min(redshift_list),
                                       vmax=np.max(redshift_list)))
        sm._A = []
        plt.colorbar(sm, label='Redshift')

        if include_err is True:
            plt.errorbar(col_x, col_y, xerr=err_x, yerr=err_y,
                         ms=2, alpha=0.5, ls=' ', zorder=0)

        return fig

    def plot_ig_space(self, test_pts, test_vals, filter_idx,
                      return_centers=False):

        """
        Plot the information gain space by filter center position.
        """

        f = filters()

        filt_centers = []

        for filter_set in test_pts:

            if self.set_ratio is not None:

                filter_info = []

                for i in range(int(len(filter_set)/2)):
                    edges = np.array(filter_set[2*i:2*(i+1)])
                    bottom_len = edges[1] - edges[0]
                    top_len = self.set_ratio*bottom_len
                    center = edges[0] + bottom_len/2.
                    top_left = center - top_len/2.
                    top_right = center + top_len/2.
                    filter_info.append([edges[0], top_left,
                                        top_right, edges[1]])
            else:
                filter_info = [filter_set[4*i:4*(i+1)]
                               for i in range(int(len(filter_set)/4))]

            filt_centers.append(self.find_filt_centers(filter_info))

        filt_centers = np.array(filt_centers)

        keep_idx = []
        for idx, filter_vals in list(enumerate(filt_centers)):
            if np.sum(1.0*np.isnan(filter_vals)) == 0:
                keep_idx.append(idx)
        filt_centers = filt_centers[keep_idx]

        xi, yi = filt_centers[:, filter_idx].T
        triang = tri.Triangulation(xi, yi)

        xx, yy = np.meshgrid(np.linspace(np.min(xi), np.max(xi), 100),
                             np.linspace(np.min(yi), np.max(yi), 100))

        interp_lin = tri.LinearTriInterpolator(triang, test_vals[keep_idx])
        zi_lin = interp_lin(xx, yy)

        extent = [np.min(xi), np.max(xi), np.min(yi), np.max(yi)]
        plt.imshow(zi_lin, cmap=plt.cm.plasma, origin='lower',
                   extent=extent, interpolation='bicubic')

        if return_centers is True:
            return xi, yi

    def plot_color_distributions(self, filter_names, redshift_list,
                                 cmap=plt.get_cmap('plasma'), fig=None,
                                 add_cbar=False):

        """
        Plot the color-color tracks for each SED template as a function
        of redshift using pairs of filters.
        """

        """
        Plot the color-color tracks for each SED template as a function
        of redshift using pairs of filters.
        """

        if fig is None:
            fig = plt.figure(figsize=(12, 6))

        shift_seds = []
        shift_values = []

        for sed_obj in self.sed_list:
            for idx, z_val in list(enumerate(redshift_list)):
                sed_copy = deepcopy(sed_obj)
                sed_copy.redshiftSED(z_val)
                shift_seds.append(sed_copy)
                shift_values.append(idx)

        color_x_dict = BandpassDict([self.filter_dict[filt] for filt
                                     in filter_names], filter_names)

        calc_ig = calcIG(color_x_dict, shift_seds,
                         np.ones(len(shift_seds)),
                         sky_mag=self.sky_mag,
                         sed_mags=self.sed_mags)
        col_x, err_x = calc_ig.calc_colors()

        num_z = len(redshift_list)

        cmap = plt.get_cmap('plasma')

        c_extent = np.linspace(0, 1, len(redshift_list))

        i = 0
        for c, err, idx in zip(col_x, err_x, shift_values):
            pts = np.linspace(c-5*err, c+5*err, 100)
            vals = stats.norm.pdf(pts, loc=c, scale=err)
            plt.fill_between(pts, 0, vals, alpha=0.3,
                             color=cmap(c_extent[idx]))
            i += 1

        plt.ylabel('Probability Density')
        plt.xlabel('%s - %s Color' % (filter_names[0],
                                      filter_names[1]))

        if add_cbar is True:
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(
                                        vmin=np.min(redshift_list),
                                        vmax=np.max(redshift_list)))
            sm._A = []

            new_ax = fig.add_axes([0.07, 0.00, 0.91, 0.03])

            cbar = plt.colorbar(sm, cax=new_ax, orientation='horizontal')
            cbar.set_label('Galaxy Redshift')

        return fig



from setuptools import setup, find_packages

setup(
    name="siggi",
    version="0.1.5",
    author="Bryce Kalmbach",
    author_email="brycek@uw.edu",
    url="https://github.com/jbkalmbach/siggi",
    packages=find_packages(),
    description="Spectral Information Gain Optimization",
    long_description=open("README.md").read(),
    package_data={"": ["README.md", "LICENSE"],
                  "siggi": ["data/*",
                            "data/lsst_baseline_throughputs/*"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        ],
)
